// compile with
// cl /nologo /LD FileCRLF.cpp /link /DEF:FileCRLF.def advapi32.lib

#include <windows.h>
#define _FAR_USE_FARFINDDATA
#define _FAR_NO_NAMELESS_UNIONS
#include "plugin.hpp"

enum
{
  MFileCase,

  MOk,
  MCancel,

  MCvtTitle,
  MProcessSubDir,

  MConverting,
  MCurRun,

};

// TYPES OF CONVERSIONS

#define MODE_TO_UNIX  0
#define MODE_TO_WIN   1

struct Options
{
  int ConvertMode;
  int ProcessSubDir;
} Opt;

struct InitDialogItem
{
  unsigned char Type;
  unsigned char X1,Y1,X2,Y2;
  unsigned char Focus;
  unsigned int Selected;
  unsigned int Flags;
  unsigned char DefaultButton;
  char *Data;
};

const char *GetMsg(int MsgId);
void InitDialogItems(const struct InitDialogItem *Init,
                     struct FarDialogItem *Item,
                     int ItemsNumber);
char *GetFullName(char *Dest,const char *Dir,char *Name);
void ProcessName(char *OldFullName, DWORD FileAttributes);


void SetRegKey(HKEY hRoot,const char *Key,const char *ValueName,DWORD ValueData);
int GetRegKey(HKEY hRoot,const char *Key,const char *ValueName,int &ValueData,DWORD Default);
int GetRegKey(HKEY hRoot,const char *Key,const char *ValueName,DWORD Default);

static struct PluginStartupInfo Info;
static struct FarStandardFunctions FSF;
char PluginRootKey[80];
BOOL IsOldFar;

const char *GetMsg(int MsgId)
{
  return(Info.GetMsg(Info.ModuleNumber,MsgId));
}

void InitDialogItems(const struct InitDialogItem *Init,struct FarDialogItem *Item,
                    int ItemsNumber)
{
  int I;
  struct FarDialogItem *PItem=Item;
  const struct InitDialogItem *PInit=Init;
  for (I=0;I<ItemsNumber;I++,PItem++,PInit++)
  {
    PItem->Type=PInit->Type;
    PItem->X1=PInit->X1;
    PItem->Y1=PInit->Y1;
    PItem->X2=PInit->X2;
    PItem->Y2=PInit->Y2;
    PItem->Focus=PInit->Focus;
    PItem->Param.Selected=PInit->Selected;
    PItem->Flags=PInit->Flags;
    PItem->DefaultButton=PInit->DefaultButton;
    lstrcpy(PItem->Data.Data,((unsigned int)PInit->Data<2000)?
           GetMsg((unsigned int)PInit->Data):PInit->Data);
  }
}

char * lstrchr(char *s, int c) 
{
  char *p;
  for (p = s ; *p ; p++)
    if (*p == c)
      return p;
  return c ? NULL : p;
}

char *lstrrchr(char *s, int c) 
{
  char *p;
  for (p = lstrchr(s, 0) ; p >= s ; p--)
    if (*p == c)
      return p;
  return c ? NULL : p;
}

char *GetOnlyName(char *FullName)
{
  char *Name=lstrrchr(FullName,'\\');
  if(Name) ++Name;
  else Name=FullName;
  return Name;
}

char *GetFullName(char *Dest,const char *Dir,char *Name)
{
  lstrcpy(Dest,Dir);
  int len=lstrlen(Dest);
  if(len)
  {
    if(Dest[len-1]=='\\') --len;
    else Dest[len]='\\';
    lstrcpy(Dest+len+1,GetOnlyName(Name));
  }
  else lstrcpy(Dest, Name);

  return Dest;
}

void CRLFConversion()
{
  struct InitDialogItem InitItems[]={
  /* 00 */DI_DOUBLEBOX,3,1,42,10,0,0,0,0,(char *)MCvtTitle,
  /* 01 */DI_RADIOBUTTON,5,2,0,0,0,0,DIF_GROUP,0,"to &Unix",
  /* 02 */DI_RADIOBUTTON,5,3,0,0,0,0,0,0,"to &Windows",
  /* 03 */DI_TEXT,5,4,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,"",
  /* 04 */DI_CHECKBOX,5,5,0,0,0,0,0,0,(char *)MProcessSubDir,
  /* 05 */DI_TEXT,5,6,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,"",
  /* 06 */DI_CHECKBOX,5,7,0,0,0,0,0,0,(char *)MCurRun,
  /* 07 */DI_TEXT,5,8,0,0,0,0,DIF_BOXCOLOR|DIF_SEPARATOR,0,"",
  /* 08 */DI_BUTTON,0,9,0,0,0,0,DIF_CENTERGROUP,1,(char *)MOk,
  /* 09 */DI_BUTTON,0,9,0,0,0,0,DIF_CENTERGROUP,0,(char *)MCancel
  };
  struct FarDialogItem DialogItems[sizeof(InitItems)/sizeof(InitItems[0])];
  InitDialogItems(InitItems,DialogItems,sizeof(InitItems)/sizeof(InitItems[0]));

  DialogItems[1+Opt.ConvertMode].Focus=DialogItems[1+Opt.ConvertMode].Param.Selected=TRUE;
  DialogItems[4].Param.Selected=Opt.ProcessSubDir;
  DialogItems[6].Param.Selected=0;

  int ExitCode=Info.Dialog(Info.ModuleNumber,-1,-1,46,12,
                           "Contents",DialogItems,
                           sizeof(DialogItems)/sizeof(DialogItems[0]));
  if (ExitCode!=8)
    return;

  int I;
  if (DialogItems[1].Param.Selected) I = MODE_TO_UNIX;   else
  if (DialogItems[2].Param.Selected) I = MODE_TO_WIN;    else
    return;

  struct Options Backup;

  if(!DialogItems[6].Param.Selected)
    Backup=Opt;

  Opt.ConvertMode=I;
  Opt.ProcessSubDir=DialogItems[4].Param.Selected;

  struct PanelInfo PInfo;
  Info.Control(INVALID_HANDLE_VALUE,FCTL_GETPANELINFO,&PInfo);

  HANDLE hScreen=Info.SaveScreen(0,0,-1,-1);
  const char *MsgItems[]={GetMsg(MCvtTitle),GetMsg(MConverting)};
  Info.Message(Info.ModuleNumber,0,NULL,MsgItems,sizeof(MsgItems)/sizeof(MsgItems[0]),0);

  char FullName[NM];

  for (I=0;I < PInfo.SelectedItemsNumber; I++)
  {
    GetFullName(FullName,PInfo.CurDir,
                PInfo.SelectedItems[I].FindData.cFileName);

    ProcessName(FullName,PInfo.SelectedItems[I].FindData.dwFileAttributes);
  }

  if(!DialogItems[6].Param.Selected)
  {
    SetRegKey(HKEY_CURRENT_USER,"","ConvertMode",Opt.ConvertMode);
    SetRegKey(HKEY_CURRENT_USER,"","ProcessSubDir",Opt.ProcessSubDir);
  }
  else
    Opt=Backup;

  Info.RestoreScreen(hScreen);
  Info.Control(INVALID_HANDLE_VALUE,FCTL_UPDATEPANEL,NULL);
  Info.Control(INVALID_HANDLE_VALUE,FCTL_REDRAWPANEL,NULL);
}

HKEY CreateRegKey(HKEY hRoot,const char *Key);
HKEY OpenRegKey(HKEY hRoot,const char *Key);

void SetRegKey(HKEY hRoot,const char *Key,const char *ValueName,DWORD ValueData)
{
  HKEY hKey=CreateRegKey(hRoot,Key);
  RegSetValueEx(hKey,ValueName,0,REG_DWORD,(BYTE *)&ValueData,sizeof(ValueData));
  RegCloseKey(hKey);
}

int GetRegKey(HKEY hRoot,const char *Key,const char *ValueName,int &ValueData,DWORD Default)
{
  HKEY hKey=OpenRegKey(hRoot,Key);
  DWORD Type,Size=sizeof(ValueData);
  int ExitCode=RegQueryValueEx(hKey,ValueName,0,&Type,(BYTE *)&ValueData,&Size);
  RegCloseKey(hKey);
  if (hKey==NULL || ExitCode!=ERROR_SUCCESS)
  {
    ValueData=Default;
    return(FALSE);
  }
  return(TRUE);
}

int GetRegKey(HKEY hRoot,const char *Key,const char *ValueName,DWORD Default)
{
  int ValueData;
  GetRegKey(hRoot,Key,ValueName,ValueData,Default);
  return(ValueData);
}

HKEY CreateRegKey(HKEY hRoot,const char *Key)
{
  HKEY hKey;
  DWORD Disposition;
  char FullKeyName[512];
  lstrcpy(FullKeyName,PluginRootKey);
  if (*Key)
  {
    lstrcat(FullKeyName,"\\");
    lstrcat(FullKeyName,Key);
  }
  RegCreateKeyEx(hRoot,FullKeyName,0,NULL,0,KEY_WRITE,NULL,
                 &hKey,&Disposition);
  return(hKey);
}

HKEY OpenRegKey(HKEY hRoot,const char *Key)
{
  HKEY hKey;
  char FullKeyName[512];
  lstrcpy(FullKeyName,PluginRootKey);
  if (*Key)
  {
    lstrcat(FullKeyName,"\\");
    lstrcat(FullKeyName,Key);
  }
  if (RegOpenKeyEx(hRoot,FullKeyName,0,KEY_QUERY_VALUE,&hKey)!=ERROR_SUCCESS)
    return(NULL);
  return(hKey);
}

char file_buffer [2048] ;
char out_buffer  [4096] ;

FILETIME ft1, ft2, ft3  ;

//   Process case convetsion of single file\directore
//     ProcessName - convert case of given filename
//                   use all options from `Opt`
//                   Call recurse for subdirectories
void ProcessName(char *OldFullName, DWORD FileAttributes)
{
  char NewFullName[NM];

  //Recurse to directories
  if( Opt.ProcessSubDir && (FileAttributes&FILE_ATTRIBUTE_DIRECTORY))
  {
    struct PluginPanelItem *PanelItems;
    int ItemsNumber,DirList;

    DirList = Info.GetDirList(OldFullName,&PanelItems,&ItemsNumber);

    if(DirList && ItemsNumber)
    {
      for (DWORD I=0; I < ItemsNumber; I++)
      {
        GetFullName(NewFullName,OldFullName,PanelItems[I].FindData.cFileName);
        ProcessName(NewFullName,PanelItems[I].FindData.dwFileAttributes);
      }
    }
    Info.FreeDirList(PanelItems);
    return ;
  }

  GetTempFileName (".", "cvt~", 0, NewFullName) ;

  HANDLE hin  = CreateFile (OldFullName, GENERIC_READ,  0, 0, OPEN_EXISTING, 0, 0) ;
  if (hin  == INVALID_HANDLE_VALUE)                       return ;

  HANDLE hout = CreateFile (NewFullName, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0, 0) ;
  if (hout == INVALID_HANDLE_VALUE) { CloseHandle (hin) ; return ; }

  DWORD remains = 0 ;
  int   still   = 1 ;

  while (still)
  {
    DWORD dw ;
    ReadFile (hin, file_buffer + remains, sizeof (file_buffer), &dw, 0) ;

    if (dw == 0) still = 0 ;

    remains += dw ;
    dw       = 0  ;

    DWORD i;
    for (i = 0 ; i < remains ; ++i)
    {
      if (file_buffer [i] == '\n')
      {
        if (i + 1 == remains && still) break ;
        else
        {
          if (file_buffer [i + 1] == '\r') ++i ;

          if (Opt.ConvertMode == MODE_TO_UNIX)
            out_buffer [dw++] = '\n' ;
          else
          {
            out_buffer [dw++] = 13   ;
            out_buffer [dw++] = 10   ;
          }

          continue ;
        }
      }

      if (file_buffer [i] == '\r')
      {
        if (i + 1 == remains && still) break ;
        else
        {
          if (file_buffer [i + 1] == '\n') ++i ;

          if (Opt.ConvertMode == MODE_TO_UNIX)
            out_buffer [dw++] = '\n' ;
          else
          {
            out_buffer [dw++] = 13   ;
            out_buffer [dw++] = 10   ;
          }

          continue ;
        }
      }

      out_buffer [dw++] = file_buffer [i] ;
    }

    WriteFile  (hout, out_buffer, dw, &dw, 0) ;
    MoveMemory (file_buffer, file_buffer + i, remains - i) ;

    remains -= i ;
  }

  if (GetFileTime (hin, &ft1, &ft2, &ft3)) SetFileTime (hout, &ft1, &ft2, &ft3) ;

  CloseHandle (hout) ;
  CloseHandle (hin)  ;
  MoveFileEx  (NewFullName, OldFullName, MOVEFILE_REPLACE_EXISTING) ;
}

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *Info)
{
  ::Info=*Info;
  IsOldFar=TRUE;
  if(Info->StructSize >= sizeof(struct PluginStartupInfo))
  {
    ::FSF=*Info->FSF;
    ::Info.FSF=&::FSF;
    IsOldFar=FALSE;

    lstrcpy(PluginRootKey,Info->RootKey);
    lstrcat(PluginRootKey,"\\CRLFConversion");

    Opt.ConvertMode=GetRegKey(HKEY_CURRENT_USER,"","ConvertMode",0);
    Opt.ProcessSubDir=GetRegKey(HKEY_CURRENT_USER,"","ProcessSubDir",0);
  }
}


HANDLE WINAPI _export OpenPlugin(int /*OpenFrom*/,int /*Item*/)
{
  if(!IsOldFar)
    CRLFConversion();
  return(INVALID_HANDLE_VALUE);
}


void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  if(!IsOldFar)
  {
    Info->StructSize=sizeof(*Info);
    static char *PluginMenuStrings[1];
    PluginMenuStrings[0]=(char*)GetMsg(MFileCase);
    Info->PluginMenuStrings=PluginMenuStrings;
    Info->PluginMenuStringsNumber=sizeof(PluginMenuStrings)/sizeof(PluginMenuStrings[0]);
  }
}
