#include <all_far.h>
#pragma hdrstop

#include "_Int.h"

static HANDLE hScreen = NULL;
static void SSave( void )
  {
    hScreen = FP_Info->SaveScreen(0,0,-1,-1);
}
static void SRestore( void )
  {
    if ( hScreen ) {
      FP_Info->RestoreScreen(hScreen);
      hScreen = NULL;
    }
}

static void EMessage( CONSTSTR Msg,int type = 0)
  {  CONSTSTR Strs[] = {
       FP_GetMsg( MError ),
       0,
       FP_GetMsg( MOk ) };
     Strs[1] = FP_GetMsg( Msg );

     if ( GetLastError() != ERROR_SUCCESS )
       type |= FMSG_ERRORTYPE;

     FP_Info->Message( 0,type,NULL,Strs,sizeof(Strs)/sizeof(Strs[0]),1 );
}

static void FRename( CONSTSTR From,CONSTSTR To )
  {  CONSTSTR Strs[] = {
       FP_GetMsg( MRename ),
       0,
       FP_GetMsg( MTo ),
       0 };
     Strs[1] = From;
     Strs[3] = To;

     static width = 0;
     int w;
     w = Max( strlen(From),strlen(To) );

     if ( w < width ) {
       SRestore();
       SSave();
     }

     FP_Info->Message( FP_Info->ModuleNumber,FMSG_LEFTALIGN,NULL,Strs,sizeof(Strs)/sizeof(Strs[0]),0 );

     width = w;
}

static int ERenameFail( CONSTSTR File,CONSTSTR To )
  {  CONSTSTR Strs[] = {
        FP_GetMsg(MError),
        FP_GetMsg(MRenameFail),
        0,
        FP_GetMsg( MTo ),
        0,
        FP_GetMsg(MSkip), FP_GetMsg(MAll), FP_GetMsg(MCancel)
     };
     Strs[2] = File;
     Strs[4] = To;

 return FP_Info->Message( 0,FMSG_WARNING|FMSG_ERRORTYPE,NULL,Strs,sizeof(Strs)/sizeof(Strs[0]),3 );
}

BOOL     Save( char *FileName,PPluginPanelItem p,int cn );
CONSTSTR Read( char *FileName,PFP_ItemList il );

#define ACT( v,msg ) if ( (v) == 0 ) { EMessage(msg,FMSG_WARNING); return; }
void DoAction( void )
  {  PanelInfo   pi;
     char        FileName[ MAX_PATH_SIZE ];
     int         n;
     CONSTSTR    m;
     FP_ItemList il;
     bool        SkipAll;

     FileName[0] = 0;

//Get files
     SetLastError(ERROR_SUCCESS);
     ACT( FP_Info->Control( INVALID_HANDLE_VALUE,FCTL_GETPANELINFO,&pi),
          FMSG(MEGetList) )

     ACT( pi.PanelType == PTYPE_FILEPANEL,
          FMSG(MEFilesPanel) )

     ACT( pi.SelectedItemsNumber > 0,
          FMSG(MENoFiles) )

//Save
     ACT( FP_FSF &&
          FP_FSF->MkTemp( FileName,"FRename" ) &&
          Save(FileName,pi.SelectedItems,pi.SelectedItemsNumber),
          FMSG(MESave) )

//Edit files list
     n = FP_Info->Editor( FileName,
                          FP_GetMsg(MRenameTitle),
                          0,0,-1,-1,EF_DISABLEHISTORY,
                          11,1 );
     //0 - file open error;
     if ( n == 0 ) {
       EMessage( FMSG(MEOpen),FMSG_WARNING );
       goto End;
     } else
     //1 - success. File was modified;
     if ( n == 1 ) {
        ;
     } else
     //2 - success. File was not modified;
     if ( n == 2 ) {
       SetLastError(ERROR_SUCCESS);
       if ( Opt.InformCancel )
       EMessage( FMSG(MECancelled) );
       goto End;
     } else
     //3 - file loading was stopped by user.
     if ( n == 3 ) {
       SetLastError(ERROR_SUCCESS);
       if ( Opt.ConfirmCancelled )
         EMessage( FMSG(MEAborted),FMSG_WARNING );
       goto End;
     }

//Read files
     SetLastError(ERROR_SUCCESS);
     ACT( (m=Read(FileName,&il)) == NULL, m )

//Rename
     SSave();

     SkipAll = false;

     for ( n = 0; n < il.Count(); n++ ) {
       char *src  = il[n]->FindData.cFileName,
            *dest = il[n]->Description;

       FRename( src,dest );

       if ( MoveFile(src,dest) || GetLastError() == ERROR_FILE_NOT_FOUND ||
            SkipAll )
         continue;

       switch( ERenameFail(src,dest) ) {
         case  0: /*Skip*/                     break;
         case  1: /*Skip all*/ SkipAll = true; break;
         case -1: /*ESC*/
         case  2: /*Cancel*/   goto Done;
       }
     }

//Say count
Done:
     SRestore();
     FP_Info->Control(INVALID_HANDLE_VALUE,FCTL_UPDATEPANEL,NULL);
     FP_Info->Control(INVALID_HANDLE_VALUE,FCTL_REDRAWPANEL,NULL);
     if ( il.Count() && Opt.ConfirmResults ) {
       SetLastError(ERROR_SUCCESS);
       char str[ MAX_PATH_SIZE ];
       SNprintf( str,sizeof(str)-1,
                 "%s %d of %d %s",
                 FP_GetMsg(MProcessd), n, il.Count(), FP_GetMsg(MFiles) );
       EMessage( str );
     }

//All done
End:
     SRestore();
     if (FileName[0])
       DeleteFile(FileName);
}

BOOL Save( char *FileName,PPluginPanelItem p,int cn )
  {  FILE *f = fopen( FileName,"w" );
     int   n,w;

     if ( !f ) return FALSE;

     fprintf( f, FP_GetMsg(MHelp1) ); fprintf( f,"\n" );
     fprintf( f, FP_GetMsg(MHelp2) ); fprintf( f,"\n\n" );
     fprintf( f, FP_GetMsg(MHelp3) ); fprintf( f,"\n" );
     fprintf( f, FP_GetMsg(MHelp4) ); fprintf( f,"\n" );
     fprintf( f, FP_GetMsg(MHelp5) ); fprintf( f,"\n\n" );
     fprintf( f, FP_GetMsg(MHelp6) ); fprintf( f,"\n\n" );
     fprintf( f, "-------------------------------------------------------------\n" );

     for ( w = n = 0; n < cn; n++ )
       w = Max( w,strLen( p[n].FindData.cFileName) );
     w += 2;

     for ( n = 0; n < cn; n++ )
       if ( fprintf( f,"%%%s%%%*c%%%s%%\n",
                     p[n].FindData.cFileName,
                     w-strLen( p[n].FindData.cFileName),' ',
                     p[n].FindData.cFileName ) <= 0 )
         return FALSE;

 return fclose(f) == 0;
}

CONSTSTR Read( char *FileName,PFP_ItemList il )
  {  FILE *f = fopen( FileName,"r" );
     char  str[ MAX_PATH_SIZE*3 ],
           fn[ MAX_PATH_SIZE ];
     char *m,*start;

     PluginPanelItem it,*p;

     if ( !f ) return FMSG(MEOpenList);

     while( !feof(f) && fgets(str,sizeof(str)-1,f) ) {
       MemSet( &it,0,sizeof(it) );
       it.Description = fn;

       for ( m = str; *m && *m != '%'; m++ );
       if ( !m[0] ) continue;

       m++;
       for( start = m; *m && *m != '%'; m++ );
       if ( !m[0] ) continue;

       StrCpy( it.FindData.cFileName,start,Min(MAX_PATH_SIZE,m-start+1) );

       m++;
       for( ; *m && *m != '%'; m++ );
       if ( !m[0] ) continue;

       m++;
       for( start = m; *m && *m != '%'; m++ );
       if ( !m[0] ) continue;

       StrCpy( it.Description,start,Min(MAX_PATH_SIZE,m-start+1) );

       if ( StrCmp(it.FindData.cFileName,it.Description) == 0 )
         continue;

       il->Add( &it );
     }

     fclose(f);

 return NULL;
}
